@@ -128,6 +128,8 @@ static int alsa_set_param(struct state *state, const char *k, const char *s)
 		state->props.use_chmap = spa_atob(s);
 	} else if (spa_streq(k, "api.alsa.multi-rate")) {
 		state->multi_rate = spa_atob(s);
+	} else if (spa_streq(k, "api.alsa.htimestamp")) {
+		state->htimestamp = spa_atob(s);
 	} else if (spa_streq(k, "latency.internal.rate")) {
 		state->process_latency.rate = atoi(s);
 	} else if (spa_streq(k, "latency.internal.ns")) {
@@ -308,6 +310,14 @@ struct spa_pod *spa_alsa_enum_propinfo(struct state *state,
 			SPA_PROP_INFO_params, SPA_POD_Bool(true));
 		break;
 	case 14:
+		param = spa_pod_builder_add_object(b,
+			SPA_TYPE_OBJECT_PropInfo, SPA_PARAM_PropInfo,
+			SPA_PROP_INFO_name, SPA_POD_String("api.alsa.htimestamp"),
+			SPA_PROP_INFO_description, SPA_POD_String("Use hires timestamps"),
+			SPA_PROP_INFO_type, SPA_POD_CHOICE_Bool(state->htimestamp),
+			SPA_PROP_INFO_params, SPA_POD_Bool(true));
+		break;
+	case 15:
 		param = spa_pod_builder_add_object(b,
 			SPA_TYPE_OBJECT_PropInfo, SPA_PARAM_PropInfo,
 			SPA_PROP_INFO_name, SPA_POD_String("latency.internal.rate"),
@@ -316,7 +326,7 @@ struct spa_pod *spa_alsa_enum_propinfo(struct state *state,
 				0, 65536),
 			SPA_PROP_INFO_params, SPA_POD_Bool(true));
 		break;
-	case 15:
+	case 16:
 		param = spa_pod_builder_add_object(b,
 			SPA_TYPE_OBJECT_PropInfo, SPA_PARAM_PropInfo,
 			SPA_PROP_INFO_name, SPA_POD_String("latency.internal.ns"),
@@ -325,7 +335,7 @@ struct spa_pod *spa_alsa_enum_propinfo(struct state *state,
 				0LL, 2 * SPA_NSEC_PER_SEC),
 			SPA_PROP_INFO_params, SPA_POD_Bool(true));
 		break;
-	case 16:
+	case 17:
 		param = spa_pod_builder_add_object(b,
 			SPA_TYPE_OBJECT_PropInfo, SPA_PARAM_PropInfo,
 			SPA_PROP_INFO_name, SPA_POD_String("clock.name"),
@@ -394,6 +404,9 @@ int spa_alsa_add_prop_params(struct state *state, struct spa_pod_builder *b)
 	spa_pod_builder_string(b, "api.alsa.multi-rate");
 	spa_pod_builder_bool(b, state->multi_rate);
 
+	spa_pod_builder_string(b, "api.alsa.htimestamp");
+	spa_pod_builder_bool(b, state->htimestamp);
+
 	spa_pod_builder_string(b, "latency.internal.rate");
 	spa_pod_builder_int(b, state->process_latency.rate);
 
@@ -477,14 +490,24 @@ static cookie_io_functions_t io_funcs = {
 	.write = log_write,
 };
 
+static void silence_error_handler(const char *file, int line,
+		const char *function, int err, const char *fmt, ...)
+{
+}
+
 int spa_alsa_init(struct state *state, const struct spa_dict *info)
 {
 	uint32_t i;
 	int err;
+	const char *str;
 
 	snd_config_update_free_global();
 
+	if ((str = spa_dict_lookup(info, "device.profile.pro")) != NULL)
+		state->is_pro = spa_atob(str);
+
 	state->multi_rate = true;
+	state->htimestamp = false;
 	for (i = 0; info && i < info->n_items; i++) {
 		const char *k = info->items[i].key;
 		const char *s = info->items[i].value;
@@ -546,6 +569,57 @@ int spa_alsa_clear(struct state *state)
 	return err;
 }
 
+static int probe_pitch_ctl(struct state *state, const char* device_name)
+{
+	snd_ctl_elem_id_t *id;
+	/* TODO: Add configuration params for the control name and units */
+	const char *elem_name =
+		state->stream == SND_PCM_STREAM_CAPTURE ?
+		"Capture Pitch 1000000" :
+		"Playback Pitch 1000000";
+	int err;
+
+	snd_lib_error_set_handler(silence_error_handler);
+
+	err = snd_ctl_open(&state->ctl, device_name, SND_CTL_NONBLOCK);
+	if (err < 0) {
+		spa_log_info(state->log, "%s could not find ctl device: %s",
+				state->props.device, snd_strerror(err));
+		state->ctl = NULL;
+		goto error;
+	}
+
+	snd_ctl_elem_id_alloca(&id);
+	snd_ctl_elem_id_set_name(id, elem_name);
+	snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_PCM);
+
+	snd_ctl_elem_value_malloc(&state->pitch_elem);
+	snd_ctl_elem_value_set_id(state->pitch_elem, id);
+
+	err = snd_ctl_elem_read(state->ctl, state->pitch_elem);
+	if (err < 0) {
+		spa_log_debug(state->log, "%s: did not find ctl %s: %s",
+				state->props.device, elem_name, snd_strerror(err));
+
+		snd_ctl_elem_value_free(state->pitch_elem);
+		state->pitch_elem = NULL;
+
+		snd_ctl_close(state->ctl);
+		state->ctl = NULL;
+		goto error;
+	}
+
+	snd_ctl_elem_value_set_integer(state->pitch_elem, 0, 1000000);
+	CHECK(snd_ctl_elem_write(state->ctl, state->pitch_elem), "snd_ctl_elem_write");
+	state->last_rate = 1.0;
+
+	spa_log_info(state->log, "%s: found ctl %s", state->props.device, elem_name);
+	err = 0;
+error:
+	snd_lib_error_set_handler(NULL);
+	return err;
+}
+
 int spa_alsa_open(struct state *state, const char *params)
 {
 	int err;
@@ -588,6 +662,8 @@ int spa_alsa_open(struct state *state, const char *params)
 	state->sample_count = 0;
 	state->sample_time = 0;
 
+	probe_pitch_ctl(state, device_name);
+
 	return 0;
 
 error_exit_close:
@@ -622,6 +698,14 @@ int spa_alsa_close(struct state *state)
 	state->have_format = false;
 	state->opened = false;
 
+	if (state->pitch_elem) {
+		snd_ctl_elem_value_free(state->pitch_elem);
+		state->pitch_elem = NULL;
+
+		snd_ctl_close(state->ctl);
+		state->ctl = NULL;
+	}
+
 	return err;
 }
 
@@ -1062,6 +1146,10 @@ static int enum_pcm_formats(struct state *state, uint32_t index, uint32_t *next,
 			}
 		}
 	}
+	if (j > 1)
+		choice->body.type = SPA_CHOICE_Enum;
+	spa_pod_builder_pop(b, &f[1]);
+
 	if (j == 0) {
 		char buf[1024];
 		int i, r, offs;
@@ -1090,9 +1178,6 @@ static int enum_pcm_formats(struct state *state, uint32_t index, uint32_t *next,
 		spa_log_warn(state->log, "%s: access:%s", state->props.device, buf);
 		return -ENOTSUP;
 	}
-	if (j > 1)
-		choice->body.type = SPA_CHOICE_Enum;
-	spa_pod_builder_pop(b, &f[1]);
 
 	if ((res = add_rate(state, 1, 1, false, index & 0xffff, next, 0, params, b)) != 1)
 		return res;
@@ -1564,31 +1649,27 @@ int spa_alsa_set_format(struct state *state, struct spa_audio_info *fmt, uint32_
 
 	dir = 0;
 	period_size = state->default_period_size;
-	is_batch = snd_pcm_hw_params_is_batch(params) &&
-		!state->disable_batch;
+	is_batch = snd_pcm_hw_params_is_batch(params) && !state->disable_batch;
 
 	/* no period size specified. If we are batch or not using timers,
 	 * use the graph duration as the period */
 	if (period_size == 0 && (is_batch || state->disable_tsched))
-		period_size = state->position ? state->position->clock.duration : DEFAULT_PERIOD;
+		period_size = state->position ? state->position->clock.target_duration : DEFAULT_PERIOD;
+	if (period_size == 0)
+		period_size = DEFAULT_PERIOD;
 
-	if (is_batch) {
-		if (period_size == 0)
-			period_size = DEFAULT_PERIOD;
-		/* batch devices get their hw pointers updated every period. Make
-		 * the period smaller and add one period of headroom. Limit the
-		 * period size to our default so that we don't create too much
-		 * headroom. */
-		if (!state->disable_tsched)
+	if (!state->disable_tsched) {
+		if (is_batch) {
+			/* batch devices get their hw pointers updated every period. Make
+			 * the period smaller and add one period of headroom. Limit the
+			 * period size to our default so that we don't create too much
+			 * headroom. */
 			period_size = SPA_MIN(period_size, DEFAULT_PERIOD) / 2;
-		spa_log_info(state->log, "%s: batch mode, period_size:%ld",
-			state->props.device, period_size);
-	} else {
-		if (period_size == 0)
-			period_size = DEFAULT_PERIOD;
-		/* disable ALSA wakeups, if we use a timer */
-		if (!state->disable_tsched && snd_pcm_hw_params_can_disable_period_wakeup(params))
-			CHECK(snd_pcm_hw_params_set_period_wakeup(hndl, params, 0), "set_period_wakeup");
+		} else {
+			/* disable ALSA wakeups */
+			if (snd_pcm_hw_params_can_disable_period_wakeup(params))
+				CHECK(snd_pcm_hw_params_set_period_wakeup(hndl, params, 0), "set_period_wakeup");
+		}
 	}
 
 	CHECK(snd_pcm_hw_params_set_period_size_near(hndl, params, &period_size, &dir), "set_period_size_near");
@@ -1648,7 +1729,7 @@ int spa_alsa_set_format(struct state *state, struct spa_audio_info *fmt, uint32_
 
 	spa_log_info(state->log, "%s (%s): format:%s access:%s-%s rate:%d channels:%d "
 			"buffer frames %lu, period frames %lu, periods %u, frame_size %zd "
-			"headroom %u start-delay:%u tsched:%u",
+			"headroom %u start-delay:%u batch:%u tsched:%u",
 			state->props.device,
 			state->stream == SND_PCM_STREAM_CAPTURE ? "capture" : "playback",
 			snd_pcm_format_name(state->format),
@@ -1656,7 +1737,7 @@ int spa_alsa_set_format(struct state *state, struct spa_audio_info *fmt, uint32_
 			planar ? "planar" : "interleaved",
 			state->rate, state->channels, state->buffer_frames, state->period_frames,
 			periods, state->frame_size, state->headroom, state->start_delay,
-			!state->disable_tsched);
+			is_batch, !state->disable_tsched);
 
 	/* write the parameters to device */
 	CHECK(snd_pcm_hw_params(hndl, params), "set_hw_params");
@@ -1664,6 +1745,41 @@ int spa_alsa_set_format(struct state *state, struct spa_audio_info *fmt, uint32_
 	return match ? 0 : 1;
 }
 
+int spa_alsa_update_rate_match(struct state *state)
+{
+	uint64_t pitch, last_pitch;
+	int err;
+
+	if (!state->pitch_elem)
+		return -ENOENT;
+
+	/* The rate/pitch defines the rate of input to output (if there were a
+	 * resampler, it's the ratio of input samples to output samples). This
+	 * means that to adjust the playback rate, we need to apply the inverse
+	 * of the given rate. */
+	if (state->stream == SND_PCM_STREAM_CAPTURE) {
+		pitch = 1000000 * state->rate_match->rate;
+		last_pitch = 1000000 * state->last_rate;
+	} else {
+		pitch = 1000000 / state->rate_match->rate;
+		last_pitch = 1000000 / state->last_rate;
+	}
+
+	/* The pitch adjustment is limited to 1 ppm */
+	if (pitch == last_pitch)
+		return 0;
+
+	snd_ctl_elem_value_set_integer(state->pitch_elem, 0, pitch);
+	CHECK(snd_ctl_elem_write(state->ctl, state->pitch_elem), "snd_ctl_elem_write");
+
+	spa_log_trace_fp(state->log, "%s %u set rate to %g",
+			state->props.device, state->stream, state->rate_match->rate);
+
+	state->last_rate = state->rate_match->rate;
+
+	return 0;
+}
+
 static int set_swparams(struct state *state)
 {
 	snd_pcm_t *hndl = state->hndl;
@@ -1815,14 +1931,18 @@ static int alsa_recover(struct state *state, int err)
 
 		delay = SPA_TIMEVAL_TO_USEC(&diff);
 		missing = delay * state->rate / SPA_USEC_PER_SEC;
+		if (missing == 0)
+			missing = state->threshold;
 
 		spa_log_trace(state->log, "%p: xrun of %"PRIu64" usec %"PRIu64,
 				state, delay, missing);
 
+		if (state->clock)
+			state->clock->xrun += missing;
+		state->sample_count += missing;
+
 		spa_node_call_xrun(&state->callbacks,
 				SPA_TIMEVAL_TO_USEC(&trigger), delay, NULL);
-
-		state->sample_count += missing ? missing : state->threshold;
 		break;
 	}
 	case SND_PCM_STATE_SUSPENDED:
@@ -1855,7 +1975,7 @@ recover:
 	return do_start(state);
 }
 
-static int get_avail(struct state *state, uint64_t current_time)
+static int get_avail(struct state *state, uint64_t current_time, snd_pcm_uframes_t *delay)
 {
 	int res, missed;
 	snd_pcm_sframes_t avail;
@@ -1864,7 +1984,7 @@ static int get_avail(struct state *state, uint64_t current_time)
 		if ((res = alsa_recover(state, avail)) < 0)
 			return res;
 		if ((avail = snd_pcm_avail(state->hndl)) < 0) {
-			if ((missed = ratelimit_test(&state->rate_limit, current_time)) >= 0) {
+			if ((missed = spa_ratelimit_test(&state->rate_limit, current_time)) >= 0) {
 				spa_log_warn(state->log, "%s: (%d missed) snd_pcm_avail after recover: %s",
 						state->props.device, missed, snd_strerror(avail));
 			}
@@ -1873,52 +1993,62 @@ static int get_avail(struct state *state, uint64_t current_time)
 	} else {
 		state->alsa_recovering = false;
 	}
-	return avail;
-}
+	*delay = avail;
 
-#if 0
-static int get_avail_htimestamp(struct state *state, uint64_t current_time)
-{
-	int res, missed;
-	snd_pcm_uframes_t avail;
-	snd_htimestamp_t tstamp;
-	uint64_t then;
+	if (state->htimestamp) {
+		snd_pcm_uframes_t havail;
+		snd_htimestamp_t tstamp;
+		uint64_t then;
 
-	if ((res = snd_pcm_htimestamp(state->hndl, &avail, &tstamp)) < 0) {
-		if ((res = alsa_recover(state, avail)) < 0)
-			return res;
-		if ((res = snd_pcm_htimestamp(state->hndl, &avail, &tstamp)) < 0) {
-			if ((missed = ratelimit_test(&state->rate_limit, current_time)) >= 0) {
+		if ((res = snd_pcm_htimestamp(state->hndl, &havail, &tstamp)) < 0) {
+			if ((missed = spa_ratelimit_test(&state->rate_limit, current_time)) >= 0) {
 				spa_log_warn(state->log, "%s: (%d missed) snd_pcm_htimestamp error: %s",
 					state->props.device, missed, snd_strerror(res));
 			}
-			avail = state->threshold * 2;
+			return avail;
 		}
-	} else {
-		state->alsa_recovering = false;
-	}
+		avail = havail;
+		*delay = havail;
+		if ((then = SPA_TIMESPEC_TO_NSEC(&tstamp)) != 0) {
+			int64_t diff;
 
-	if ((then = SPA_TIMESPEC_TO_NSEC(&tstamp)) != 0) {
-		if (then < current_time)
-			avail += (current_time - then) * state->rate / SPA_NSEC_PER_SEC;
-		else
-			avail -= (then - current_time) * state->rate / SPA_NSEC_PER_SEC;
+			if (then < current_time)
+				diff = ((int64_t)(current_time - then)) * state->rate / SPA_NSEC_PER_SEC;
+			else
+				diff = -((int64_t)(then - current_time)) * state->rate / SPA_NSEC_PER_SEC;
+
+			spa_log_trace_fp(state->log, "%"PRIu64" %"PRIu64" %"PRIi64, current_time, then, diff);
+
+			if (SPA_ABS(diff) < state->threshold * 3) {
+				*delay += SPA_CLAMP(diff, -((int64_t)state->threshold), (int64_t)state->threshold);
+				state->htimestamp_error = 0;
+			} else {
+				if (++state->htimestamp_error > MAX_HTIMESTAMP_ERROR) {
+					spa_log_error(state->log, "%s: wrong htimestamps from driver, disabling",
+						state->props.device);
+					state->htimestamp_error = 0;
+					state->htimestamp = false;
+				}
+				else if ((missed = spa_ratelimit_test(&state->rate_limit, current_time)) >= 0) {
+					spa_log_warn(state->log, "%s: (%d missed) impossible htimestamp diff:%"PRIi64,
+						state->props.device, missed, diff);
+				}
+			}
+		}
 	}
-	return SPA_MIN(avail, state->buffer_frames);
+	return avail;
 }
-#endif
 
-static int get_status(struct state *state, uint64_t current_time,
+static int get_status(struct state *state, uint64_t current_time, snd_pcm_uframes_t *avail,
 		snd_pcm_uframes_t *delay, snd_pcm_uframes_t *target)
 {
-	int avail;
+	int res;
+	snd_pcm_uframes_t a, d;
 
-	if ((avail = get_avail(state, current_time)) < 0)
-		return avail;
+	if ((res = get_avail(state, current_time, &d)) < 0)
+		return res;
 
-	avail = SPA_MIN(avail, (int)state->buffer_frames);
-
-	*target = state->threshold + state->headroom;
+	a = SPA_MIN(res, (int)state->buffer_frames);
 
 	if (state->resample && state->rate_match) {
 		state->delay = state->rate_match->delay;
@@ -1927,12 +2057,14 @@ static int get_status(struct state *state, uint64_t current_time,
 		state->delay = 0;
 		state->read_size = state->threshold;
 	}
-
 	if (state->stream == SND_PCM_STREAM_PLAYBACK) {
-		*delay = state->buffer_frames - avail;
+		*avail = state->buffer_frames - a;
+		*delay = state->buffer_frames - SPA_MIN(d, state->buffer_frames);
+		*target = state->threshold + state->headroom;
 	} else {
-		*delay = avail;
-		*target = SPA_MAX(*target, state->read_size + state->headroom);
+		*avail = a;
+		*delay = d;
+		*target = SPA_MAX(state->threshold, state->read_size) + state->headroom;
 	}
 	*target = SPA_CLAMP(*target, state->min_delay, state->max_delay);
 	return 0;
@@ -2003,7 +2135,10 @@ static int update_time(struct state *state, uint64_t current_time, snd_pcm_sfram
 		else
 			state->rate_match->rate = 1.0/corr;
 
-		SPA_FLAG_UPDATE(state->rate_match->flags, SPA_IO_RATE_MATCH_FLAG_ACTIVE, state->matching);
+		if (state->pitch_elem && state->matching)
+			spa_alsa_update_rate_match(state);
+		else
+			SPA_FLAG_UPDATE(state->rate_match->flags, SPA_IO_RATE_MATCH_FLAG_ACTIVE, state->matching);
 	}
 
 	state->next_time += state->threshold / corr * 1e9 / state->rate;
@@ -2043,7 +2178,7 @@ static int setup_matching(struct state *state)
 	if (spa_streq(state->position->clock.name, state->clock_name))
 		state->matching = false;
 
-	state->resample = ((uint32_t)state->rate != state->rate_denom) || state->matching;
+	state->resample = !state->pitch_elem && (((uint32_t)state->rate != state->rate_denom) || state->matching);
 
 	spa_log_info(state->log, "driver clock:'%s'@%d our clock:'%s'@%d matching:%d resample:%d",
 			state->position->clock.name, state->rate_denom,
@@ -2104,11 +2239,11 @@ int spa_alsa_write(struct state *state)
 
 	if (state->following && state->alsa_started) {
 		uint64_t current_time;
-		snd_pcm_uframes_t delay, target;
+		snd_pcm_uframes_t avail, delay, target;
 
 		current_time = state->position->clock.nsec;
 
-		if (SPA_UNLIKELY((res = get_status(state, current_time, &delay, &target)) < 0))
+		if (SPA_UNLIKELY((res = get_status(state, current_time, &avail, &delay, &target)) < 0))
 			return res;
 
 		if (SPA_UNLIKELY((res = update_time(state, current_time, delay, target, true)) < 0))
@@ -2122,17 +2257,18 @@ int spa_alsa_write(struct state *state)
 			else
 				lev = SPA_LOG_LEVEL_INFO;
 
-			if ((missed = ratelimit_test(&state->rate_limit, current_time)) >= 0) {
-				spa_log_lev(state->log, lev, "%s: follower delay:%ld target:%ld thr:%u, "
-						"resync (%d missed)", state->props.device, delay,
+			if ((missed = spa_ratelimit_test(&state->rate_limit, current_time)) >= 0) {
+				spa_log_lev(state->log, lev, "%s: follower avail:%lu delay:%ld "
+						"target:%ld thr:%u, resync (%d missed)",
+						state->props.device, avail, delay,
 						target, state->threshold, missed);
 			}
 
-			if (delay > target)
-				snd_pcm_rewind(state->hndl, delay - target);
-			else if (delay < target)
-				spa_alsa_silence(state, target - delay);
-			delay = target;
+			if (avail > target)
+				snd_pcm_rewind(state->hndl, avail - target);
+			else if (avail < target)
+				spa_alsa_silence(state, target - avail);
+			avail = target;
 			state->alsa_sync = false;
 		} else
 			state->alsa_sync_warning = true;
@@ -2146,6 +2282,7 @@ again:
 		if (SPA_UNLIKELY((res = snd_pcm_mmap_begin(hndl, &my_areas, &offset, &frames)) < 0)) {
 			spa_log_error(state->log, "%s: snd_pcm_mmap_begin error: %s",
 					state->props.device, snd_strerror(res));
+			alsa_recover(state, res);
 			return res;
 		}
 		spa_log_trace_fp(state->log, "%p: begin offset:%ld avail:%ld threshold:%d",
@@ -2343,11 +2480,9 @@ int spa_alsa_read(struct state *state)
 
 		current_time = state->position->clock.nsec;
 
-		if ((res = get_status(state, current_time, &delay, &target)) < 0)
+		if ((res = get_status(state, current_time, &avail, &delay, &target)) < 0)
 			return res;
 
-		avail = delay;
-
 		if (SPA_UNLIKELY((res = update_time(state, current_time, delay, target, true)) < 0))
 			return res;
 
@@ -2359,17 +2494,18 @@ int spa_alsa_read(struct state *state)
 			else
 				lev = SPA_LOG_LEVEL_INFO;
 
-			if ((missed = ratelimit_test(&state->rate_limit, current_time)) >= 0) {
+			if ((missed = spa_ratelimit_test(&state->rate_limit, current_time)) >= 0) {
 				spa_log_lev(state->log, lev, "%s: follower delay:%ld target:%ld thr:%u, "
 						"resync (%d missed)", state->props.device, delay,
 						target, state->threshold, missed);
 			}
 
-			if (delay < target)
-				max_read = target - delay;
-			else if (delay > target)
-				snd_pcm_forward(state->hndl, delay - target);
-			delay = target;
+			if (avail < target)
+				max_read = target - avail;
+			else if (avail > target) {
+				snd_pcm_forward(state->hndl, avail - target);
+				avail = target;
+			}
 			state->alsa_sync = false;
 		} else
 			state->alsa_sync_warning = true;
@@ -2385,6 +2521,7 @@ int spa_alsa_read(struct state *state)
 		if ((res = snd_pcm_mmap_begin(hndl, &my_areas, &offset, &to_read)) < 0) {
 			spa_log_error(state->log, "%s: snd_pcm_mmap_begin error: %s",
 					state->props.device, snd_strerror(res));
+			alsa_recover(state, res);
 			return res;
 		}
 		spa_log_trace_fp(state->log, "%p: begin offs:%ld frames:%ld to_read:%ld thres:%d", state,
@@ -2457,13 +2594,15 @@ int spa_alsa_skip(struct state *state)
 }
 
 
-static int handle_play(struct state *state, uint64_t current_time,
+static int handle_play(struct state *state, uint64_t current_time, snd_pcm_uframes_t avail,
 		snd_pcm_uframes_t delay, snd_pcm_uframes_t target)
 {
+	struct spa_io_buffers *io = state->io;
 	int res;
 
 	if (state->alsa_started && SPA_UNLIKELY(delay > target + state->max_error)) {
-		spa_log_trace(state->log, "%p: early wakeup %lu %lu", state, delay, target);
+		spa_log_trace(state->log, "%p: early wakeup %ld %lu %lu", state,
+				avail, delay, target);
 		if (delay > target * 3)
 			delay = target * 3;
 		state->next_time = current_time + (delay - target) * SPA_NSEC_PER_SEC / state->rate;
@@ -2473,32 +2612,24 @@ static int handle_play(struct state *state, uint64_t current_time,
 	if (SPA_UNLIKELY((res = update_time(state, current_time, delay, target, false)) < 0))
 		return res;
 
-	if (spa_list_is_empty(&state->ready)) {
-		struct spa_io_buffers *io = state->io;
+	spa_log_trace_fp(state->log, "%p: %d", state, io->status);
 
-		spa_log_trace_fp(state->log, "%p: %d", state, io->status);
+	update_sources(state, false);
 
-		update_sources(state, false);
-
-		io->status = SPA_STATUS_NEED_DATA;
-		res = spa_node_call_ready(&state->callbacks, SPA_STATUS_NEED_DATA);
-	}
-	else {
-		res = spa_alsa_write(state);
-	}
-	return res;
+	io->status = SPA_STATUS_NEED_DATA;
+	return spa_node_call_ready(&state->callbacks, SPA_STATUS_NEED_DATA);
 }
 
-static int handle_capture(struct state *state, uint64_t current_time,
+static int handle_capture(struct state *state, uint64_t current_time, snd_pcm_uframes_t avail,
 		snd_pcm_uframes_t delay, snd_pcm_uframes_t target)
 {
 	int res;
 	struct spa_io_buffers *io;
 
-	if (SPA_UNLIKELY(delay < state->read_size)) {
-		spa_log_trace(state->log, "%p: early wakeup %ld %ld %d", state, delay, target,
-				state->read_size);
-		state->next_time = current_time + (target - delay) * SPA_NSEC_PER_SEC /
+	if (SPA_UNLIKELY(avail < state->read_size)) {
+		spa_log_trace(state->log, "%p: early wakeup %ld %ld %ld %d", state,
+				delay, avail, target, state->read_size);
+		state->next_time = current_time + (state->read_size - avail) * SPA_NSEC_PER_SEC /
 			state->rate;
 		return -EAGAIN;
 	}
@@ -2543,9 +2674,9 @@ static uint64_t get_time_ns(struct state *state)
 static void alsa_wakeup_event(struct spa_source *source)
 {
 	struct state *state = source->data;
-	snd_pcm_uframes_t delay, target;
+	snd_pcm_uframes_t avail, delay, target;
 	uint64_t expire, current_time;
-	int res;
+	int res, missed;
 
 	if (SPA_UNLIKELY(state->disable_tsched)) {
 		/* ALSA poll fds need to be "demangled" to know whether it's a real wakeup */
@@ -2593,7 +2724,7 @@ static void alsa_wakeup_event(struct spa_source *source)
 		return;
 	}
 
-	if (SPA_UNLIKELY(get_status(state, current_time, &delay, &target) < 0)) {
+	if (SPA_UNLIKELY(get_status(state, current_time, &avail, &delay, &target) < 0)) {
 		spa_log_error(state->log, "get_status error");
 		state->next_time += state->threshold * 1e9 / state->rate;
 		goto done;
@@ -2602,24 +2733,28 @@ static void alsa_wakeup_event(struct spa_source *source)
 #ifndef FASTPATH
 	if (SPA_UNLIKELY(spa_log_level_topic_enabled(state->log, SPA_LOG_TOPIC_DEFAULT, SPA_LOG_LEVEL_TRACE))) {
 		uint64_t nsec = get_time_ns(state);
-		spa_log_trace_fp(state->log, "%p: wakeup %lu %lu %"PRIu64" %"PRIu64" %"PRIi64
-				" %d %"PRIi64, state, delay, target, nsec, nsec,
+		spa_log_trace_fp(state->log, "%p: wakeup %lu %lu %lu %"PRIu64" %"PRIu64" %"PRIi64
+				" %d %"PRIi64, state, avail, delay, target, nsec, nsec,
 				nsec - current_time, state->threshold, state->sample_count);
 	}
 #endif
 
 	if (state->stream == SND_PCM_STREAM_PLAYBACK)
-		handle_play(state, current_time, delay, target);
+		handle_play(state, current_time, avail, delay, target);
 	else
-		handle_capture(state, current_time, delay, target);
+		handle_capture(state, current_time, avail, delay, target);
 
 done:
 	if (!state->disable_tsched &&
 			(state->next_time > current_time + SPA_NSEC_PER_SEC ||
 			 current_time > state->next_time + SPA_NSEC_PER_SEC)) {
-		spa_log_error(state->log, "%s: impossible timeout %lu %lu %"PRIu64" %"PRIu64" %"PRIi64
-				" %d %"PRIi64, state->props.device, delay, target, current_time, state->next_time,
-				state->next_time - current_time, state->threshold, state->sample_count);
+		if ((missed = spa_ratelimit_test(&state->rate_limit, current_time)) >= 0) {
+			spa_log_error(state->log, "%s: impossible timeout %lu %lu %lu %"
+					PRIu64" %"PRIu64" %"PRIi64" %d %"PRIi64" (%d missed)",
+					state->props.device, avail, delay, target,
+					current_time, state->next_time, state->next_time - current_time,
+					state->threshold, state->sample_count, missed);
+		}
 		state->next_time = current_time + state->threshold * 1e9 / state->rate;
 	}
 	set_timeout(state, state->next_time);
